function plugindef()
	-- This function and the 'finaleplugin' namespace
	-- are both reserved for the plug-in definition.
	finaleplugin.MinJWLuaVersion = "0.45"
	finaleplugin.Author = "Chris"
	finaleplugin.Copyright = "Copyright (c) Chris, 2019"
	finaleplugin.Version = "2.0-beta"
	finaleplugin.RevisionNotes = [[
refonte totale du plugin
]]
	finaleplugin.Date = "14/11/2019"
	finaleplugin.AuthorEmail = "chris@finale-logiciel-aide-gravure-musicale.eu"
	finaleplugin.Id = "chris.copyvertsmartshapes"
	finaleplugin.CategoryTags = "Region, Smartshape"
	return "Copie verticale de lignes avancées", "", ""
end

--[[
 VARIABLES GLOBALES (paramètres par défaut)
]]
gFl = nil -- langue/language  -> français : gFl = 1 / anglais : gFl = 2
gSelectedSmartShapes = 4  -- smart shapes à copier (1 bit par smart shape)
gAttachement = 0 -- 0 -> Tout / 1 -> Attachées aux notes / 2 -> Attachées aux temps
gExceptEmpty = true  -- sauf les mesures vides
gNormalStaff = true  -- copie sur portées normales
gPercussionStaff = true  -- copie sur portées percussions
gTablatureStaff = true  -- copie sur tablatures


--[[
 CONSTANTES
]]
-- tous les textes de l'interface
local INTERFACE_TEXTS = {
	pluginName = {"Copie verticale de lignes avancées " .. finaleplugin.Version,"Vertical copy of smart shapes " .. finaleplugin.Version},
	slur =  {"Liaison","Slur"},
	slurDashed =  {"Liaison pointillée","Dashed slur"},
	hairpin = {"Crescendo","Hairpin"},
	doubleEndedBracket = {"Crochet double","Double-Ended Bracket"},
	dashedDoubleEndedBracket = {"Crochet double pointillé","Dashed Double-Ended Bracket"},
	bracket = {"Crochet","Bracket"},
	dashedBracket = {"Crochet pointillé","Dashed Bracket"},
	line = {"Ligne","Line"},
	dashLine = {"Ligne pointillée","Dashed Line"},
	dashCurve = {"Courbe pointillée","Dashed Curve"},
	customLine = {"Ligne personnalisée","Custom Line"},
	colDlItems = {"Lignes avancées","Smart Shapes"},
	btAllSmartShapes = {"Toutes","All"},
	btNoneSmartShapes = {"Aucune","None"},
	lblAttachment = {"Attachement :","Attachment :"},
	popupAttached = {{"Toutes","Attachées aux notes","Attachées aux temps"},{"All","Note attached","Beat attached"}},
	cbEmpty = {"Ne pas copier sur mesures vides","Do not copy on empty measures"},
	colDlStaves = {"Portées","Staves"},
	btAllStaves = {"Toutes","All"},
	btNoneStaves = {"Aucune","None"},
	cbNormalStaff = {"Portée normale","Normal Staff"},
	cbPercStaff = {"Portée percussion","Percussion Staff"},
	cbTabStaff = {"Tablature","Tablature"},
	cbPreview = {"Aperçu","Preview"},
	lblThanks = {"JW Lua " .. finenv.StringVersion .. " - Merci à Jari Williamsson","JW Lua " .. finenv.StringVersion .. " - Thanks to Jari Williamsson"},
	btPrefs = {"Définir par défaut","Set as default"},
	btCancel = {"Annuler","Cancel"},
	msgPrefsOK =  {"Vos paramètres par défaut ont bien été sauvegardés","Your default settings have been saved"},
	msgPrefsError =  {"Une erreur s'est produite. Vos paramètres par défaut n'ont pu être sauvegardés","An error has occurred. Your default settings could not be saved"},
	msgSelection =  {"Veuillez effectuer une sélection dans une seule portée.","Please make a selection in a single staff."}
}


-- slurs auto -> attached notes
local SHAPES_TYPES = {
	{name = function() return INTERFACE_TEXTS.slur[gFl] end, shapeType = {"SMARTSHAPE_SLURAUTO","SMARTSHAPE_SLURUP","SMARTSHAPE_SLURDOWN"}},
	{name = function() return INTERFACE_TEXTS.slurDashed[gFl] end, shapeType = {"SMARTSHAPE_DASHEDSLURAUTO","SMARTSHAPE_DASHEDSLURUP","SMARTSHAPE_DASHEDSLURDOWN"}},
	{name = function() return INTERFACE_TEXTS.hairpin[gFl] end, shapeType = {"SMARTSHAPE_CRESCENDO","SMARTSHAPE_DIMINUENDO"}},
	{name = function() return INTERFACE_TEXTS.doubleEndedBracket[gFl] end, shapeType = {"SMARTSHAPE_SOLIDLINEDOWN2","SMARTSHAPE_SOLIDLINEUP2"}},
	{name = function() return INTERFACE_TEXTS.dashedDoubleEndedBracket[gFl] end, shapeType = {"SMARTSHAPE_DASHLINEDOWN2","SMARTSHAPE_DASHLINEUP2"}},
	{name = function() return INTERFACE_TEXTS.bracket[gFl] end, shapeType = {"SMARTSHAPE_SOLIDLINEDOWN","SMARTSHAPE_SOLIDLINEDOWNLEFT ","SMARTSHAPE_SOLIDLINEUP","SMARTSHAPE_SOLIDLINEUPLEFT","SMARTSHAPE_SOLIDLINEUPDOWN","SMARTSHAPE_SOLIDLINEDOWNUP"}},
	{name = function() return INTERFACE_TEXTS.dashedBracket[gFl] end, shapeType = {"SMARTSHAPE_DASHLINEUP","SMARTSHAPE_DASHLINEDOWN","SMARTSHAPE_DASHLINEDOWNLEFT","SMARTSHAPE_DASHLINEUPLEFT","SMARTSHAPE_DASHLINEUPDOWN","SMARTSHAPE_DASHLINEDOWNUP"}},
	{name = function() return INTERFACE_TEXTS.line[gFl] end, shapeType = {"SMARTSHAPE_SOLIDLINE"}},
	{name = function() return INTERFACE_TEXTS.dashLine[gFl] end, shapeType = {"SMARTSHAPE_DASHLINE"}},
	{name = function() return INTERFACE_TEXTS.dashCurve[gFl] end, shapeType = {"SMARTSHAPE_DASHCURVEAUTO","SMARTSHAPE_DASHCURVEUP","SMARTSHAPE_DASHCURVEDOWN"}},
	{name = function() return INTERFACE_TEXTS.customLine[gFl] end, shapeType = {"SMARTSHAPE_CUSTOM"}}
}


--[[
***** FONCTIONS UTILITAIRES *****
]]
local function createString(texte)
	local str=finale.FCString()
	str.LuaString = texte
	return str
end
-- convertit une Table indexée en FCStrings
local function tableToStrings(table)
	local strings =finale.FCStrings()
	for _, v in ipairs(table) do
		strings:AddCopy(createString(v))
	end
	return strings
end
--[[
définit la langue (fr -> 0 , en -> 1) 
]]
local function getLanguage()
	if gFl then return end
	local categorydefs = finale.FCCategoryDefs()
	categorydefs:LoadAll()
	for cd in each(categorydefs) do
		local name = cd:CreateName()
		if name.LuaString == "Repères" then
			gFl = 1
			return
		end
	end
	gFl = 2
end

local function ConvertUtf8ToUtf16(utf8str)
	local unicode_to_1252 = {}
	unicode_to_1252[0x0000]=0x00         unicode_to_1252[0x0001]=0x01        unicode_to_1252[0x0002]=0x02        unicode_to_1252[0x0003]=0x03        unicode_to_1252[0x0004]=0x04
	unicode_to_1252[0x0005]=0x05        unicode_to_1252[0x0006]=0x06        unicode_to_1252[0x0007]=0x07        unicode_to_1252[0x0008]=0x08        unicode_to_1252[0x0009]=0x09
	unicode_to_1252[0x000A]=0x0A        unicode_to_1252[0x000B]=0x0B        unicode_to_1252[0x000C]=0x0C        unicode_to_1252[0x000D]=0x0D        unicode_to_1252[0x000E]=0x0E
	unicode_to_1252[0x000F]=0x0F        unicode_to_1252[0x0010]=0x10        unicode_to_1252[0x0011]=0x11        unicode_to_1252[0x0012]=0x12        unicode_to_1252[0x0013]=0x13
	unicode_to_1252[0x0014]=0x14        unicode_to_1252[0x0015]=0x15        unicode_to_1252[0x0016]=0x16        unicode_to_1252[0x0017]=0x17        unicode_to_1252[0x0018]=0x18
	unicode_to_1252[0x0019]=0x19        unicode_to_1252[0x001A]=0x1A        unicode_to_1252[0x001B]=0x1B        unicode_to_1252[0x001C]=0x1C        unicode_to_1252[0x001D]=0x1D
	unicode_to_1252[0x001E]=0x1E        unicode_to_1252[0x001F]=0x1F        unicode_to_1252[0x0020]=0x20        unicode_to_1252[0x0021]=0x21        unicode_to_1252[0x0022]=0x22
	unicode_to_1252[0x0023]=0x23        unicode_to_1252[0x0024]=0x24        unicode_to_1252[0x0025]=0x25        unicode_to_1252[0x0026]=0x26        unicode_to_1252[0x0027]=0x27
	unicode_to_1252[0x0028]=0x28        unicode_to_1252[0x0029]=0x29        unicode_to_1252[0x002A]=0x2A        unicode_to_1252[0x002B]=0x2B        unicode_to_1252[0x002C]=0x2C
	unicode_to_1252[0x002D]=0x2D        unicode_to_1252[0x002E]=0x2E        unicode_to_1252[0x002F]=0x2F        unicode_to_1252[0x0030]=0x30        unicode_to_1252[0x0031]=0x31
	unicode_to_1252[0x0032]=0x32        unicode_to_1252[0x0033]=0x33        unicode_to_1252[0x0034]=0x34        unicode_to_1252[0x0035]=0x35        unicode_to_1252[0x0036]=0x36
	unicode_to_1252[0x0037]=0x37        unicode_to_1252[0x0038]=0x38        unicode_to_1252[0x0039]=0x39        unicode_to_1252[0x003A]=0x3A        unicode_to_1252[0x003B]=0x3B
	unicode_to_1252[0x003C]=0x3C        unicode_to_1252[0x003D]=0x3D        unicode_to_1252[0x003E]=0x3E        unicode_to_1252[0x003F]=0x3F        unicode_to_1252[0x0040]=0x40
	unicode_to_1252[0x0041]=0x41        unicode_to_1252[0x0042]=0x42        unicode_to_1252[0x0043]=0x43        unicode_to_1252[0x0044]=0x44        unicode_to_1252[0x0045]=0x45
	unicode_to_1252[0x0046]=0x46        unicode_to_1252[0x0047]=0x47        unicode_to_1252[0x0048]=0x48        unicode_to_1252[0x0049]=0x49        unicode_to_1252[0x004A]=0x4A
	unicode_to_1252[0x004B]=0x4B        unicode_to_1252[0x004C]=0x4C        unicode_to_1252[0x004D]=0x4D        unicode_to_1252[0x004E]=0x4E        unicode_to_1252[0x004F]=0x4F
	unicode_to_1252[0x0050]=0x50        unicode_to_1252[0x0051]=0x51        unicode_to_1252[0x0052]=0x52        unicode_to_1252[0x0053]=0x53        unicode_to_1252[0x0054]=0x54
	unicode_to_1252[0x0055]=0x55        unicode_to_1252[0x0056]=0x56        unicode_to_1252[0x0057]=0x57        unicode_to_1252[0x0058]=0x58        unicode_to_1252[0x0059]=0x59
	unicode_to_1252[0x005A]=0x5A        unicode_to_1252[0x005B]=0x5B        unicode_to_1252[0x005C]=0x5C        unicode_to_1252[0x005D]=0x5D        unicode_to_1252[0x005E]=0x5E
	unicode_to_1252[0x005F]=0x5F        unicode_to_1252[0x0060]=0x60        unicode_to_1252[0x0061]=0x61        unicode_to_1252[0x0062]=0x62        unicode_to_1252[0x0063]=0x63
	unicode_to_1252[0x0064]=0x64        unicode_to_1252[0x0065]=0x65        unicode_to_1252[0x0066]=0x66        unicode_to_1252[0x0067]=0x67        unicode_to_1252[0x0068]=0x68
	unicode_to_1252[0x0069]=0x69        unicode_to_1252[0x006A]=0x6A        unicode_to_1252[0x006B]=0x6B        unicode_to_1252[0x006C]=0x6C        unicode_to_1252[0x006D]=0x6D
	unicode_to_1252[0x006E]=0x6E        unicode_to_1252[0x006F]=0x6F        unicode_to_1252[0x0070]=0x70        unicode_to_1252[0x0071]=0x71        unicode_to_1252[0x0072]=0x72
	unicode_to_1252[0x0073]=0x73        unicode_to_1252[0x0074]=0x74        unicode_to_1252[0x0075]=0x75        unicode_to_1252[0x0076]=0x76        unicode_to_1252[0x0077]=0x77
	unicode_to_1252[0x0078]=0x78        unicode_to_1252[0x0079]=0x79        unicode_to_1252[0x007A]=0x7A        unicode_to_1252[0x007B]=0x7B        unicode_to_1252[0x007C]=0x7C
	unicode_to_1252[0x007D]=0x7D        unicode_to_1252[0x007E]=0x7E        unicode_to_1252[0x007F]=0x7F        unicode_to_1252[0x20AC]=0x80        unicode_to_1252[0x201A]=0x82
	unicode_to_1252[0x0192]=0x83        unicode_to_1252[0x201E]=0x84        unicode_to_1252[0x2026]=0x85        unicode_to_1252[0x2020]=0x86        unicode_to_1252[0x2021]=0x87
	unicode_to_1252[0x02C6]=0x88        unicode_to_1252[0x2030]=0x89        unicode_to_1252[0x0160]=0x8A        unicode_to_1252[0x2039]=0x8B        unicode_to_1252[0x0152]=0x8C
	unicode_to_1252[0x017D]=0x8E        unicode_to_1252[0x2018]=0x91        unicode_to_1252[0x2019]=0x92        unicode_to_1252[0x201C]=0x93        unicode_to_1252[0x201D]=0x94
	unicode_to_1252[0x2022]=0x95        unicode_to_1252[0x2013]=0x96        unicode_to_1252[0x2014]=0x97        unicode_to_1252[0x02DC]=0x98        unicode_to_1252[0x2122]=0x99
	unicode_to_1252[0x0161]=0x9A        unicode_to_1252[0x203A]=0x9B        unicode_to_1252[0x0153]=0x9C        unicode_to_1252[0x017E]=0x9E        unicode_to_1252[0x0178]=0x9F
	unicode_to_1252[0x00A0]=0xA0        unicode_to_1252[0x00A1]=0xA1        unicode_to_1252[0x00A2]=0xA2        unicode_to_1252[0x00A3]=0xA3        unicode_to_1252[0x00A4]=0xA4
	unicode_to_1252[0x00A5]=0xA5        unicode_to_1252[0x00A6]=0xA6        unicode_to_1252[0x00A7]=0xA7        unicode_to_1252[0x00A8]=0xA8        unicode_to_1252[0x00A9]=0xA9
	unicode_to_1252[0x00AA]=0xAA        unicode_to_1252[0x00AB]=0xAB        unicode_to_1252[0x00AC]=0xAC        unicode_to_1252[0x00AD]=0xAD        unicode_to_1252[0x00AE]=0xAE
	unicode_to_1252[0x00AF]=0xAF        unicode_to_1252[0x00B0]=0xB0        unicode_to_1252[0x00B1]=0xB1        unicode_to_1252[0x00B2]=0xB2        unicode_to_1252[0x00B3]=0xB3
	unicode_to_1252[0x00B4]=0xB4        unicode_to_1252[0x00B5]=0xB5        unicode_to_1252[0x00B6]=0xB6        unicode_to_1252[0x00B7]=0xB7        unicode_to_1252[0x00B8]=0xB8
	unicode_to_1252[0x00B9]=0xB9        unicode_to_1252[0x00BA]=0xBA        unicode_to_1252[0x00BB]=0xBB        unicode_to_1252[0x00BC]=0xBC        unicode_to_1252[0x00BD]=0xBD
	unicode_to_1252[0x00BE]=0xBE        unicode_to_1252[0x00BF]=0xBF        unicode_to_1252[0x00C0]=0xC0        unicode_to_1252[0x00C1]=0xC1        unicode_to_1252[0x00C2]=0xC2
	unicode_to_1252[0x00C3]=0xC3        unicode_to_1252[0x00C4]=0xC4        unicode_to_1252[0x00C5]=0xC5        unicode_to_1252[0x00C6]=0xC6        unicode_to_1252[0x00C7]=0xC7
	unicode_to_1252[0x00C8]=0xC8        unicode_to_1252[0x00C9]=0xC9        unicode_to_1252[0x00CA]=0xCA        unicode_to_1252[0x00CB]=0xCB        unicode_to_1252[0x00CC]=0xCC
	unicode_to_1252[0x00CD]=0xCD        unicode_to_1252[0x00CE]=0xCE        unicode_to_1252[0x00CF]=0xCF        unicode_to_1252[0x00D0]=0xD0        unicode_to_1252[0x00D1]=0xD1
	unicode_to_1252[0x00D2]=0xD2        unicode_to_1252[0x00D3]=0xD3        unicode_to_1252[0x00D4]=0xD4        unicode_to_1252[0x00D5]=0xD5        unicode_to_1252[0x00D6]=0xD6
	unicode_to_1252[0x00D7]=0xD7        unicode_to_1252[0x00D8]=0xD8        unicode_to_1252[0x00D9]=0xD9        unicode_to_1252[0x00DA]=0xDA        unicode_to_1252[0x00DB]=0xDB
	unicode_to_1252[0x00DC]=0xDC        unicode_to_1252[0x00DD]=0xDD        unicode_to_1252[0x00DE]=0xDE        unicode_to_1252[0x00DF]=0xDF        unicode_to_1252[0x00E0]=0xE0
	unicode_to_1252[0x00E1]=0xE1        unicode_to_1252[0x00E2]=0xE2        unicode_to_1252[0x00E3]=0xE3        unicode_to_1252[0x00E4]=0xE4        unicode_to_1252[0x00E5]=0xE5
	unicode_to_1252[0x00E6]=0xE6        unicode_to_1252[0x00E7]=0xE7        unicode_to_1252[0x00E8]=0xE8        unicode_to_1252[0x00E9]=0xE9        unicode_to_1252[0x00EA]=0xEA
	unicode_to_1252[0x00EB]=0xEB        unicode_to_1252[0x00EC]=0xEC        unicode_to_1252[0x00ED]=0xED        unicode_to_1252[0x00EE]=0xEE        unicode_to_1252[0x00EF]=0xEF
	unicode_to_1252[0x00F0]=0xF0        unicode_to_1252[0x00F1]=0xF1        unicode_to_1252[0x00F2]=0xF2        unicode_to_1252[0x00F3]=0xF3        unicode_to_1252[0x00F4]=0xF4
	unicode_to_1252[0x00F5]=0xF5        unicode_to_1252[0x00F6]=0xF6        unicode_to_1252[0x00F7]=0xF7        unicode_to_1252[0x00F8]=0xF8        unicode_to_1252[0x00F9]=0xF9
	unicode_to_1252[0x00FA]=0xFA        unicode_to_1252[0x00FB]=0xFB        unicode_to_1252[0x00FC]=0xFC        unicode_to_1252[0x00FD]=0xFD        unicode_to_1252[0x00FE]=0xFE
	unicode_to_1252[0x00FF]=0xFF

	local undefined = ('?'):byte()
	local pos, result = 1, {}
	while pos <= #utf8str do
		local code, size = utf8str:byte(pos, pos), 1
		if code >= 0xC0 and code < 0xFE then
			local mask = 64
			code = code - 128
			repeat
				local next_byte = utf8str:byte(pos+size, pos+size) or 0
				if next_byte >= 0x80 and next_byte < 0xC0 then
					code, size = (code - mask - 2) * 64 + next_byte, size+1
				else
					code, size = utf8str:byte(pos, pos), 1
				end
				mask = mask * 32
			until code < mask
		end
		pos = pos + size
		table.insert(result, string.char(unicode_to_1252[code] or undefined))
	end
	return table.concat(result)
end

-- teste si le fichier existe
local function file_exists(filename)
	local f = io.open(ConvertUtf8ToUtf16(filename), "r")
	if f ~= nil then
		io.close(f)
		return true
	end

	return false
end

-- FONCTIONS SUR LES BITS (1er bit -> 0)

-- exemple : bit32.extract( 8, 3, 1) retourne le (3° para) bit 3 (2° para) de 8
local function hasbit(x, p)
	return bit32.extract(x, p, 1) ~= 0      
end

-- exemple : bit32.replace(1, 1, 4, 1)  écrit le bit 4 (3° para) dans 1 (1° para)
local function setbit(x, p)
	return bit32.replace(x, 1, p, 1)
end

-- bit32.replace(1, 0, 4, 1)  supprime le bit 4 (3° para) dans 1 (1° para)
local function clearbit(x, p)
	return bit32.replace(x, 0, p, 1)
end
--[[
***** FIN FONCTIONS UTILITAIRES *****
]]


--[[
FONCTIONS PARAMETRES PAR DEFAUT
]]
-- retourne le nom complet du fichier de préférences
local function getFilePrefs()
	local pathPrefs = finale.FCString()
	pathPrefs:SetPreferencesFolderPath() 
	pathPrefs:AssureEndingPathDelimiter() 
	local filename = pathPrefs.LuaString .. "chrispluginprefs.txt"
	return ConvertUtf8ToUtf16(filename)
end

-- retourne le nombre de paras chargés
local function loadPrefs()
	local filename = getFilePrefs()
	if not file_exists(filename) then return 0 end

	local strPluginID = "[" .. finaleplugin.Id .. "]"
	local isPluginID = false
	local nLoadParas = 0

	for line in io.lines(ConvertUtf8ToUtf16(filename)) do
		if string.find(line,"%[") then -- one Plug-in ID
			if isPluginID then break end  -- on a parcouru le bloc des prefs de notre plugin
			if line == strPluginID then isPluginID = true end

		elseif isPluginID then -- Paramètre
			if pcall(load(line)) then nLoadParas = nLoadParas + 1 end
		end
	end

	return nLoadParas
end

local function savePrefs(prefs)
	local filename = getFilePrefs()
	local fileprefs
	local mode = "w"
	table.insert(prefs,1,"[" .. finaleplugin.Id .. "]")  -- ajouter l'entête aux  prefs

	if file_exists(filename) then
		-- ajouter les datas existentes du fichier (sauf ceux de ce script) aux prefs
		local strPluginID = "[" .. finaleplugin.Id .. "]"
		local isPluginID = false

		for line in io.lines(filename) do
			if string.find(line,"%[") then -- one Plug-in ID
				if line == strPluginID then
					isPluginID = true 
				else
					if isPluginID then isPluginID = false end
					table.insert(prefs,line)
				end
			elseif not isPluginID then -- Paramètres des autres scripts
				table.insert(prefs,line)
			end
		end

		-- réécrire les datas et ajouter les préférences de ce script
		mode = "w+"
	end

	-- créer le fichier de préférences ou ajouter les préférences
	fileprefs = io.open(filename, mode)
	if fileprefs == nil then return false end
	for _, v in ipairs(prefs) do
		fileprefs:write(v, "\n")
	end
	fileprefs:flush()
	fileprefs:close()
	return true
end
--[[
FIN - FONCTIONS PARAMETRES PAR DEFAUT
]]

--[[
CLASSE : BOITE DE DIALOGUE
]]
local DialogBoxClass = {} -- the table representing the class, which will double as the metatable for the instances
DialogBoxClass.__index = DialogBoxClass

-- ***** CONSTRUCTEUR *****
function DialogBoxClass:new(datas, events)
	self = setmetatable({}, DialogBoxClass)

	self.dialogRef = finale.FCCustomLuaWindow()
	self.controlNames = {}

	-- création des controles
	for _, properties in ipairs(datas) do
		self:createControl(properties)
	end

	-- création des buttons OK et Cancel
	self.controlNames["btOK"] = self.dialogRef:CreateOkButton()
	self.controlNames["btCancel"] = self.dialogRef:CreateCancelButton()

	-- événements
	for evt, callback in pairs(events) do
		self.dialogRef["Register" .. evt](self.dialogRef, callback) 
	end

	return self
end

-- ***** PROPRIETES *****

-- retourne la référence de la boite de dialogue
function DialogBoxClass:getDialogRef()
	return self.dialogRef
end

-- retourne la référence d'un controle
function DialogBoxClass:getCtrl(name)
	return self.controlNames[name]
end

-- retourne tous les controleq
function DialogBoxClass:getCtrls()
	return self.controlNames
end


-- ***** METHODES *****

-- affiche la boite de dialogue
function DialogBoxClass:display(parent, callbackCancelButtonPressed)
	local ret = self:getDialogRef():ExecuteModal(parent)
	if ret == finale.EXECMODAL_CLOSE and callbackCancelButtonPressed then
		callbackCancelButtonPressed()
	end
	return ret == finale.EXECMODAL_OK
end

function DialogBoxClass:initDataList(properties)
	local datalist = self:getCtrl(properties.name)
	if not datalist then return end
	local columns = properties.arguments.columns 
	if not columns then return end

	for _, column  in pairs(columns) do
		if type(column[1]) == "string" then column[1] = createString(column[1]) end
		datalist:AddColumn(column[1],column[2])
	end

	local rows = properties.arguments.rows 
	if not rows then return end

	for  _, row in ipairs(rows) do
		local dataListRow = datalist:CreateRow()
		dataListRow:ClearAll()	-- effacer la collection pour pouvoir rajouter des FCString()
		if type(row) ~= "table" then row = {row} end   -- 1 seule colonne

		for  _, item in ipairs(row) do	-- colonnes
			if type(item) ~= "userdata" then item = createString(item) end
			dataListRow:AddCopy(item)
		end
	end
end

function DialogBoxClass:createControl(properties)
	local method = self:getDialogRef()[properties.fn]
	if not method then -- ce n'est pas une méthode de FCCustomLuaWindow
		method = self[properties.fn]
		if method then -- c'est une méthode de DialogBoxClass
			method(self, properties)
		else -- ce n'est pas une méthode de FCCustomLuaWindow, ni de DialogBoxClass
			print(properties.fn .. " n'est pas une méthode de FCCustomLuaWindow, ni de DialogBoxClass")
		end
		return
	end

	-- la majorité des méthodes de création de controle, possède 2 arguments : left et top
	-- seules "CreateHorizontalLine" et "CreateVerticalLine" en ont 3
	-- le 3° argument est donc mis à à zéro et est redéfini après (top et left probablement)
	local control = method(self:getDialogRef(), 0, 0, 0)

	for prop, value  in pairs(properties) do
		if control[prop] then    -- si la méthode existe
			local arguments = {}

			if type(value) ~= "table" then  -- un seul argument
				value = {value} -- convertir en table
			end

			-- parcourir tous les arguments
			for _, argument in ipairs(value) do
				if type(argument) == "table" then -- convertir en FCStrings
					table.insert(arguments, tableToStrings(argument))

				elseif type(argument) == "string" then
					if string.sub(argument, 1, 9) == "pControl-" then	-- controle
						table.insert(arguments, self:getCtrl(string.sub(argument, 10)))
					else
						table.insert(arguments, createString(argument))
					end
				else	-- argument autre que table et string
					table.insert(arguments, argument)
				end
			end

			-- si appel avec plus d'arguments, pas de problème (mais moins si)
			-- maximum de 3 arguments
			control[prop](control, arguments[1], arguments[2], arguments[3])

		end
	end

	if properties.name then self.controlNames[properties.name] = control end  -- mémoriser le nom du controle s'il existe
end

--[[
FIN CLASSE : BOITE DE DIALOGUE
]]
-- retourne la note liée au segment (EntryBased = true)
local function getNoteEntrySegment(seg,idStaff)
	local noteentrycell = finale.FCNoteEntryCell(seg.Measure, idStaff)
	noteentrycell:Load()
	local noteentry = noteentrycell:FindEntryNumber(seg.EntryNumber)
	return noteentry
end

-- retourne la note de la portée "copie", si elle existe à la position définie dans le segment
local function getNoteEntry(noteentryseg, seg, idStaff)
	local noteentrycell = finale.FCNoteEntryCell(seg.Measure, idStaff)
	noteentrycell:Load()

	for noteentry in each(noteentrycell) do
		if noteentry.MeasurePos == noteentryseg.MeasurePos then
			return noteentry
		end
	end

	return nil
end

local function defSegment(smartshape, seg, staffSourceID, staffDestID)
	seg:SetStaff(staffDestID)

	if smartshape.EntryBased then
		local noteentryseg =  getNoteEntrySegment(seg, staffSourceID)
		if not noteentryseg then return end

		local noteentry = getNoteEntry(noteentryseg, seg, staffDestID)
		if noteentry then seg:SetEntry(noteentry) end
	end
end

local function createSmartShapeOneStaff(smartshape, staffSourceID, staffDestID)
	local segLeft = smartshape:GetTerminateSegmentLeft()
	local segRight = smartshape:GetTerminateSegmentRight()

	defSegment(smartshape, segLeft, staffSourceID, staffDestID)
	defSegment(smartshape, segRight , staffSourceID, staffDestID)

	smartshape:SaveNewEverything(nil,nil)
end

local function isEmptyMusicRegion(musicregion,staffID)
	local musicregionDest = finale.FCMusicRegion()
	musicregionDest:SetRegion(musicregion)  -- copie de la région
	musicregionDest.StartStaff = staffID
	musicregionDest.EndStaff = staffID

	for m, s in eachcell(musicregionDest) do
		local noteentrycell = finale.FCNoteEntryCell(m, s)
		noteentrycell:Load()
		if not noteentrycell:IsEmpty() then return false end
	end
	return true
end

local function getRegion()
	local musicregion = finale.FCMusicRegion()
	musicregion:SetCurrentSelection()

	if musicregion:IsEmpty() or musicregion:CalcStaffSpan() ~= 1 then
		finenv.UI():AlertInfo(INTERFACE_TEXTS.msgSelection[gFl], INTERFACE_TEXTS.pluginName[gFl])
		return nil
	end

	return musicregion
end

local function setCheckDataList(dl, checkstate)
	local n = dl:GetCount()
	if n == 0 then return end

	for ligne = 0, n - 1 do
		local dataListRow = dl:GetItemAt(ligne)
		dataListRow:SetCheck(checkstate)  -- génère event
	end
end

local function initDatalistItems(dl)
	dl:DeleteAll()

	for ind, shapeType in ipairs(SHAPES_TYPES) do
		local dataListRow = dl:CreateRow()
		dataListRow:GetItemAt(0).LuaString = shapeType.name()

		local isSelected = hasbit(gSelectedSmartShapes, ind - 1)
		dataListRow:SetCheck(isSelected)
	end
end

-- écrit tous les textes de l'interface dans la langue choisie
local function setInterfaceNames(dialogClass)
	dialogClass:getDialogRef():SetTitle(createString(INTERFACE_TEXTS.pluginName[gFl]))

	local allControls = dialogClass:getCtrls()

	for name, ctrl in pairs(allControls) do
		if INTERFACE_TEXTS[name] then
			local typeData = type(INTERFACE_TEXTS[name][gFl]) 

			if typeData == "string" then
				ctrl:SetText(createString(INTERFACE_TEXTS[name][gFl]))

			elseif typeData == "table" then  -- poppup ou list
				local selectedItem = ctrl:GetSelectedItem()
				if selectedItem == - 1 then
					selectedItem = INTERFACE_TEXTS[name][3]()
				end
				ctrl:Clear()
				ctrl:SetStrings(tableToStrings(INTERFACE_TEXTS[name][gFl]))
				ctrl:SetSelectedItem(selectedItem)
			end
		end
	end

	initDatalistItems(dialogClass:getCtrl("dlItems"))
end

-- renvoie true si la smartshape correspond au choix de l'utilisateur
local function isUserSmartShapeSelected(smartshape)
	if gAttachement > 0 then
		if gAttachement == 1 then  -- attachées aux  notes
			if not smartshape.EntryBased then return false end
		else -- attachées aux  temps
			if smartshape.EntryBased then return false end
		end

	end
	-- chercher le type de smartshape
	for ind, v in ipairs(SHAPES_TYPES) do
		for _, v in ipairs(v.shapeType) do
			if smartshape.ShapeType == finale[v] then
				if hasbit(gSelectedSmartShapes,ind - 1) then
					return true
				else
					return false
				end
			end
		end
	end
	return false
end

local function isStaffVisible(datasStaff)
	return (datasStaff.type == finale.STAFFNOTATION_NORMAL and gNormalStaff) or
	(datasStaff.type == finale.STAFFNOTATION_PERCUSSION and gPercussionStaff) or
	(datasStaff.type == finale.STAFFNOTATION_TABLATURE  and gTablatureStaff)
end

local function updateListStaves(dl,listStaves)
	dl:DeleteAll()

	for _, datasStaff in ipairs(listStaves) do
		if datasStaff.visible then
			local dataListRow = dl:CreateRow()
			dataListRow:GetItemAt(0).LuaString = datasStaff.name
			dataListRow:GetItemAt(1):SetInteger(datasStaff.ItemNo)
			dataListRow:SetCheck(datasStaff.checked)
		end
	end
end

--[[
initialise la liste des portées avec FCSystemStaves-ScrollView (au lieu de FCStaves)
pour avoir les portées dans l'ordre du score et ne pas afficher les portées supprimées
]]
local function getListAllStaves(musicregion)
	local list = {}

	local sysstaves = finale.FCSystemStaves()
	sysstaves:LoadScrollView()
	local staff = finale.FCStaff()

	for sysstaff in each(sysstaves) do
		if staff:Load(sysstaff.Staff) then
			if staff.ItemNo ~= musicregion.StartStaff then
				local datasOneStaff = {}
				local displayName = staff:CreateDisplayFullNameString()
				if displayName:IsEmpty() then 
					displayName.LuaString = "[Staff" .. staff.ItemNo .. "]"
				end
				datasOneStaff.name = displayName.LuaString
				datasOneStaff.ItemNo = staff.ItemNo
				datasOneStaff.type = staff.NotationStyle
				datasOneStaff.checked = true
				datasOneStaff.visible = isStaffVisible(datasOneStaff)
				table.insert(list, datasOneStaff)
			end
		end
	end
	return list
end


local function apply(musicregion, listStaves, textUndo)
	finenv.StartNewUndoBlock(textUndo,false)

	local staves = finale.FCStaves()
	staves:LoadAll()
	local marks = finale.FCSmartShapeMeasureMarks()  -- (attached notes et beats)
	marks:LoadAllForRegion(musicregion, true)  -- ne charge que les marques de début

	for mark in each(marks) do
		local smartshape  = mark:CreateSmartShape(true)

		if isUserSmartShapeSelected(smartshape) then
			-- parcourir toutes les portées visibles et sélectionnées
			for _, datasStaff in ipairs(listStaves) do
				if datasStaff.visible and datasStaff.checked then
					if not gExceptEmpty or (gExceptEmpty and (not isEmptyMusicRegion(musicregion, datasStaff.ItemNo))) then
						smartshape  = mark:CreateSmartShape(true)  -- créer une SmartShape par portée
						createSmartShapeOneStaff(smartshape, musicregion.StartStaff, datasStaff.ItemNo)
					end
				end
			end
		end

	end

	finenv.UI():RedrawDocument()
end

local function createDialog(musicregion)
	local dialogClass
	local textUndo = "[JW Lua]"
	local listStaves = getListAllStaves(musicregion)
	local noEvtDataListCheck = false
	local popupAttachedSelectedItem

	local datas = {
		{fn = "CreateDataList", name = "dlItems", SetTop = 20, SetWidth = 200, SetHeight = 250, SetUseCheckboxes = true},
		{fn = "initDataList", name = "dlItems", arguments = {columns = {{INTERFACE_TEXTS.colDlItems[gFl], 200}}}},
		{fn = "CreateButton", name = "btAllSmartShapes",  SetTop = 275, SetWidth = 80},
		{fn = "CreateButton", name = "btNoneSmartShapes", SetLeft = 120, SetTop = 275, SetWidth = 80},
		{fn = "CreateStatic", name = "lblAttachment", SetTop = 320, SetWidth = 200},
		{fn = "CreatePopup", name = "popupAttached", SetTop = 340, SetWidth = 200, SetStrings = {INTERFACE_TEXTS.popupAttached[gFl]}, SetSelectedItem = gAttachement},
		{fn = "CreateCheckbox", name = "cbEmpty", SetTop = 380, SetWidth = 200, SetCheck = gExceptEmpty and 1 or 0},
		{fn = "CreateDataList", name = "dlStaves", SetLeft = 240, SetTop = 20, SetWidth = 200, SetHeight = 250, SetUseCheckboxes = true},
		{fn = "initDataList", name = "dlStaves", arguments = {columns = {{INTERFACE_TEXTS.colDlStaves[gFl], 200}, {"ID", 0}}}},
		{fn = "CreateButton", name = "btAllStaves", SetLeft = 240, SetTop = 275, SetWidth = 80},
		{fn = "CreateButton", name = "btNoneStaves", SetLeft = 360, SetTop = 275, SetWidth = 80},
		{fn = "CreateCheckbox", name = "cbNormalStaff", SetLeft = 240, SetTop = 310, SetWidth = 200, SetCheck = gNormalStaff and 1 or 0},
		{fn = "CreateCheckbox", name = "cbPercStaff", SetLeft = 240, SetTop = 330, SetWidth = 200, SetCheck = gPercussionStaff and 1 or 0},
		{fn = "CreateCheckbox", name = "cbTabStaff", SetLeft = 240, SetTop = 350, SetWidth = 200, SetCheck = gTablatureStaff and 1 or 0},
		{fn = "CreateCheckbox", name = "cbPreview", SetLeft = 200, SetTop = 440, SetWidth = 90},
		{fn = "CreateStatic", name = "lblThanks", SetLeft = 145, SetTop = 500, SetWidth = 280},
		{fn = "CreatePopup", name = "popupLanguage", SetWidth = 70, SetStrings = {{"Français","English"}}, SetSelectedItem = gFl - 1},
		{fn = "CreateButton", name = "btPrefs", SetWidth = 120}
	}

	local setCheckDataListStaves = function(checkstate)
		noEvtDataListCheck = true
		setCheckDataList(dialogClass:getCtrl("dlStaves"), checkstate)
		for _, datasStaff in ipairs(listStaves) do
			if datasStaff.visible then datasStaff.checked = checkstate end
		end
		noEvtDataListCheck = false
	end

	local setCheckDataListItems = function(checkstate)
		noEvtDataListCheck = true
		-- un event "DataListCheck" sera généré par "setCheckDataList"
		setCheckDataList(dialogClass:getCtrl("dlItems"), checkstate)
		gSelectedSmartShapes = checkstate and 2047 or 0
		noEvtDataListCheck = false
	end

	-- ***** EVENEMENTS *****
	local callbackCommand = function(ctrl)
		local ctrlID = ctrl:GetControlID()
		local cbPreview = dialogClass:getCtrl("cbPreview")
		local cbNormalStaff = dialogClass:getCtrl("cbNormalStaff")
		local cbPercStaff = dialogClass:getCtrl("cbPercStaff")
		local cbTabStaff = dialogClass:getCtrl("cbTabStaff")
			
		if ctrlID == dialogClass:getCtrl("popupLanguage"):GetControlID() then
			local value = dialogClass:getCtrl("popupLanguage"):GetSelectedItem() + 1
			if value ~= gFl then
				gFl = value
				setInterfaceNames(dialogClass)
			end
			return

		elseif ctrlID == dialogClass:getCtrl("btAllSmartShapes"):GetControlID() then
			setCheckDataListItems(true)

		elseif ctrlID == dialogClass:getCtrl("btNoneSmartShapes"):GetControlID() then
			setCheckDataListItems(false)

		elseif ctrlID == dialogClass:getCtrl("btAllStaves"):GetControlID() then
			setCheckDataListStaves(true)

		elseif ctrlID == dialogClass:getCtrl("btNoneStaves"):GetControlID() then
			setCheckDataListStaves(false)

		elseif ctrlID == dialogClass:getCtrl("popupAttached"):GetControlID() then
			gAttachement = dialogClass:getCtrl("popupAttached"):GetSelectedItem()
			if gAttachement == popupAttachedSelectedItem then return end
			popupAttachedSelectedItem = gAttachement

		elseif ctrlID == dialogClass:getCtrl("cbEmpty"):GetControlID() then
			gExceptEmpty = dialogClass:getCtrl("cbEmpty"):GetCheck() == 1 

			-- choix des types de portée
		elseif ctrlID == cbNormalStaff:GetControlID() or ctrlID == cbPercStaff:GetControlID() or ctrlID == cbTabStaff:GetControlID() then
			gNormalStaff = cbNormalStaff:GetCheck() == 1
			gPercussionStaff = cbPercStaff:GetCheck() == 1
			gTablatureStaff = cbTabStaff:GetCheck() == 1

			for _, datasStaff in ipairs(listStaves) do
				datasStaff.visible = isStaffVisible(datasStaff)
			end
			updateListStaves(dialogClass:getCtrl("dlStaves"), listStaves)

		elseif ctrlID == dialogClass:getCtrl("btPrefs"):GetControlID() then
			local prefs = {}
			table.insert(prefs,"gFl=" .. gFl)
			table.insert(prefs,"gSelectedSmartShapes=" .. gSelectedSmartShapes)
			table.insert(prefs,"gAttachement=" .. gAttachement)
			table.insert(prefs,"gExceptEmpty=" .. tostring(gExceptEmpty))
			table.insert(prefs,"gNormalStaff=" .. tostring(gNormalStaff))
			table.insert(prefs,"gPercussionStaff=" .. tostring(gPercussionStaff))
			table.insert(prefs,"gTablatureStaff=" .. tostring(gTablatureStaff))

			local msg
			if savePrefs(prefs) then
				msg = "msgPrefsOK"
			else
				msg = "msgPrefsError"
			end
			finenv.UI():AlertInfo(INTERFACE_TEXTS[msg][gFl], INTERFACE_TEXTS.pluginName[gFl])

			return
		end


		if ctrlID == cbPreview:GetControlID() then
			if cbPreview:GetCheck() == 1 then
				apply(musicregion, listStaves, textUndo) 
			else
				finenv.StartNewUndoBlock(textUndo,false)
				finenv.UI():RedrawDocument()
			end

		else    -- tous les controles sauf cbPreview, OK, Cancel
			if cbPreview:GetCheck() == 1 then
				apply(musicregion, listStaves, textUndo)
			end
		end

	end

	-- lineindex base 0
	local callbackDataListCheck = function(ctrl, lineindex, checkstate)
		if noEvtDataListCheck then return end
		if lineindex == -1 then return end
		local ctrlID = ctrl:GetControlID()
		local dlItems = dialogClass:getCtrl("dlItems")
		local dlStaves = dialogClass:getCtrl("dlStaves")

		if ctrlID == dlItems:GetControlID() then
			if checkstate then
				gSelectedSmartShapes = setbit(gSelectedSmartShapes, lineindex)
			else
				gSelectedSmartShapes = clearbit(gSelectedSmartShapes, lineindex)
			end

		elseif ctrlID == dlStaves:GetControlID() then
			local dataListRow = dlStaves:GetItemAt(lineindex)
			local ItemNo = dataListRow:GetItemAt(1):GetInteger(0)

			for _, datasStaff in ipairs(listStaves) do
				if ItemNo == datasStaff.ItemNo then
					datasStaff.checked = checkstate
					break
				end
			end
		end

		if dialogClass:getCtrl("cbPreview"):GetCheck() == 1 then
			apply(musicregion, listStaves, textUndo)
		end
	end

	local callbackOkButtonPressed = function()
		if dialogClass:getCtrl("cbPreview"):GetCheck() == 1 then finenv.StartNewUndoBlock(textUndo,false) end
		apply(musicregion, listStaves, textUndo)
	end

	local callbackCancelButtonPressed = function()
		if dialogClass:getCtrl("cbPreview"):GetCheck() == 1 then finenv.StartNewUndoBlock(textUndo,false) end
	end
	-- ***** FIN EVENEMENTS *****

	-- définir les événements
	local events = {}
	events.HandleCommand = callbackCommand
	events.HandleDataListCheck = callbackDataListCheck
	events.HandleOkButtonPressed = callbackOkButtonPressed
	events.HandleCancelButtonPressed = callbackCancelButtonPressed

	dialogClass = DialogBoxClass:new(datas, events)

	local initWindow = function()
		noEvtDataListCheck = true
		
		local popupLanguage = dialogClass:getCtrl("popupLanguage")
		popupLanguage:SetLeft(11)
		popupLanguage:SetTop(547) -- ajustement = + 1 (???)
		local btPrefs = dialogClass:getCtrl("btPrefs")
		btPrefs:SetLeft(91)
		btPrefs:SetTop(546)
--		local btOK = dialogClass:getCtrl("btOK")
--		btOK:SetLeft(324)
--		btOK:SetTop(546)
--		local btCancel = dialogClass:getCtrl("btCancel")
--		btCancel:SetLeft(391)
--		btCancel:SetTop(546)

		popupAttachedSelectedItem = gAttachement
		updateListStaves(dialogClass:getCtrl("dlStaves"), listStaves)
		setInterfaceNames(dialogClass)
		
		noEvtDataListCheck = false
	end

	initWindow()
	dialogClass:display(nil, callbackCancelButtonPressed)
end

local function main()
	loadPrefs()
	getLanguage()
	local musicregion = getRegion()
	if not musicregion then return end

	createDialog(musicregion)
end

-- DEBUT DU PROGRAMME
main()